/*
 * kgraft_patch_bsc1012183
 *
 * Fix for CVE-2016-9555, bsc#1012183
 *
 *  Upstream commit:
 *  bf911e985d6b ("sctp: validate chunk len before actually using it")
 *
 *  SLE12(-SP1) commit:
 *  6bffdf27f35548775c9ecdc20fd3b436276f1b4c
 *
 *  SLE12-SP2 commit:
 *  44ec2400b88d232eedbe18695aa9796677aa3f34
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>

#include <net/sctp/sm.h>

#define KGR_PATCHED_MODULE      "sctp"

static sctp_disposition_t (*kgr_sctp_sf_abort_violation)(struct net *net, const struct sctp_endpoint *ep, const struct sctp_association *asoc, void *arg, sctp_cmd_seq_t *commands, const __u8 *payload, const size_t paylen);
static sctp_disposition_t (*kgr_sctp_sf_shut_8_4_5)(struct net *net, const struct sctp_endpoint *ep, const struct sctp_association *asoc, const sctp_subtype_t type, void *arg, sctp_cmd_seq_t *commands);
static sctp_disposition_t (*kgr_sctp_sf_tabort_8_4_8)(struct net *net, const struct sctp_endpoint *ep, const struct sctp_association *asoc, const sctp_subtype_t type, void *arg, sctp_cmd_seq_t *commands);
static sctp_disposition_t (*kgr_sctp_sf_pdiscard)(struct net *net, const struct sctp_endpoint *ep, const struct sctp_association *asoc, const sctp_subtype_t type, void *arg, sctp_cmd_seq_t *commands);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "sctp_sf_abort_violation", (void *)&kgr_sctp_sf_abort_violation },
	{ "sctp_sf_shut_8_4_5", (void *)&kgr_sctp_sf_shut_8_4_5 },
	{ "sctp_sf_tabort_8_4_8", (void *)&kgr_sctp_sf_tabort_8_4_8 },
	{ "sctp_sf_pdiscard", (void *)&kgr_sctp_sf_pdiscard },
};

/* inlined */
static sctp_disposition_t kgr_sctp_sf_violation_chunklen(
				     struct net *net,
				     const struct sctp_endpoint *ep,
				     const struct sctp_association *asoc,
				     const sctp_subtype_t type,
				     void *arg,
				     sctp_cmd_seq_t *commands)
{
	static const char err_str[]="The following chunk had invalid length:";

	return kgr_sctp_sf_abort_violation(net, ep, asoc, arg, commands, err_str,
					sizeof(err_str));
}

/* patched function */
sctp_disposition_t kgr_sctp_sf_ootb(struct net *net,
				const struct sctp_endpoint *ep,
				const struct sctp_association *asoc,
				const sctp_subtype_t type,
				void *arg,
				sctp_cmd_seq_t *commands)
{
	struct sctp_chunk *chunk = arg;
	struct sk_buff *skb = chunk->skb;
	sctp_chunkhdr_t *ch;
	sctp_errhdr_t *err;
	__u8 *ch_end;
	int ootb_shut_ack = 0;
	int ootb_cookie_ack = 0;

	SCTP_INC_STATS(net, SCTP_MIB_OUTOFBLUES);

	ch = (sctp_chunkhdr_t *) chunk->chunk_hdr;
	do {
		/* Report violation if the chunk is less then minimal */
		if (ntohs(ch->length) < sizeof(sctp_chunkhdr_t))
			return kgr_sctp_sf_violation_chunklen(net, ep, asoc, type, arg,
						  commands);

		/*
		 * Fix CVE-2016-9555
		 *  +6 lines
		 */
		 /* Report violation if chunk len overflows */
		 ch_end = ((__u8 *)ch) + WORD_ROUND(ntohs(ch->length));
		 if (ch_end > skb_tail_pointer(skb))
			return kgr_sctp_sf_violation_chunklen(net, ep, asoc, type, arg,
						  commands);

		 /* Now that we know we at least have a chunk header,
		 * do things that are type appropriate.
		 */
		if (SCTP_CID_SHUTDOWN_ACK == ch->type)
			ootb_shut_ack = 1;

		/* RFC 2960, Section 3.3.7
		 *   Moreover, under any circumstances, an endpoint that
		 *   receives an ABORT  MUST NOT respond to that ABORT by
		 *   sending an ABORT of its own.
		 */
		if (SCTP_CID_ABORT == ch->type)
			return kgr_sctp_sf_pdiscard(net, ep, asoc, type, arg, commands);

		/* RFC 8.4, 7) If the packet contains a "Stale cookie" ERROR
		 * or a COOKIE ACK the SCTP Packet should be silently
		 * discarded.
		 */

		if (SCTP_CID_COOKIE_ACK == ch->type)
			ootb_cookie_ack = 1;

		if (SCTP_CID_ERROR == ch->type) {
			sctp_walk_errors(err, ch) {
				if (SCTP_ERROR_STALE_COOKIE == err->cause) {
					ootb_cookie_ack = 1;
					break;
				}
			}
		}

		/*
		 * Fix CVE-2016-9555
		 *  -Report violation if chunk len overflows *
		 *  -ch_end = ((__u8 *)ch) + WORD_ROUND(ntohs(ch->length));
		 *  -if (ch_end > skb_tail_pointer(skb))
		 *  -	return kgr_sctp_sf_violation_chunklen(net, ep, asoc, type, arg,
		 *  -				  commands);
		 *  -
		 */
		ch = (sctp_chunkhdr_t *) ch_end;
	} while (ch_end < skb_tail_pointer(skb));

	if (ootb_shut_ack)
		return kgr_sctp_sf_shut_8_4_5(net, ep, asoc, type, arg, commands);
	else if (ootb_cookie_ack)
		return kgr_sctp_sf_pdiscard(net, ep, asoc, type, arg, commands);
	else
		return kgr_sctp_sf_tabort_8_4_8(net, ep, asoc, type, arg, commands);
}

static int kgr_patch_bsc1012183_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1012183_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1012183_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1012183_module_nb = {
	.notifier_call = kgr_patch_bsc1012183_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1012183_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1012183_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1012183_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1012183_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1012183_module_nb);
}
